@@ -38,19 +38,26 @@
 namespace Plasma
 {
 
-
-QHash<QString, FrameData *> FrameSvgPrivate::s_sharedFrames;
+QHash<ThemePrivate *, QHash<QString, FrameData *> > FrameSvgPrivate::s_sharedFrames;
 
 // Any attempt to generate a frame whose width or height is larger than this
 // will be rejected
 static const int MAX_FRAME_SIZE = 100000;
 
+FrameData::~FrameData()
+{
+    foreach (FrameSvg *frame, references.keys()) {
+        frame->d->frames.remove(prefix);
+    }
+}
+
+
 FrameSvg::FrameSvg(QObject *parent)
     : Svg(parent),
       d(new FrameSvgPrivate(this))
 {
     connect(this, SIGNAL(repaintNeeded()), this, SLOT(updateNeeded()));
-    d->frames.insert(QString(), new FrameData(this));
+    d->frames.insert(QString(), new FrameData(this, QString()));
 }
 
 FrameSvg::~FrameSvg()
@@ -72,7 +79,7 @@ void FrameSvg::setImagePath(const QString &path)
         // we're the only user of it, let's remove it from the shared keys
         // we don't want to deref it, however, as we'll still be using it
         const QString oldKey = d->cacheId(fd, d->prefix);
-        FrameSvgPrivate::s_sharedFrames.remove(oldKey);
+        FrameSvgPrivate::s_sharedFrames[theme()->d].remove(oldKey);
     } else {
         // others are using this frame, so deref it for ourselves
         fd->deref(this);
@@ -85,7 +92,7 @@ void FrameSvg::setImagePath(const QString &path)
         // we need to replace our frame, start by looking in the frame cache
         FrameData *oldFd = d->frames[d->prefix];
         const QString key = d->cacheId(oldFd, d->prefix);
-        fd = FrameSvgPrivate::s_sharedFrames.value(key);
+        fd = FrameSvgPrivate::s_sharedFrames[theme()->d].value(key);
 
         if (fd) {
             // we found one, so ref it and use it; we also don't need to (or want to!)
@@ -106,7 +113,7 @@ void FrameSvg::setImagePath(const QString &path)
     if (updateNeeded) {
         // ensure our frame is in the cache
         const QString key = d->cacheId(fd, d->prefix);
-        FrameSvgPrivate::s_sharedFrames.insert(key, fd);
+        FrameSvgPrivate::s_sharedFrames[theme()->d].insert(key, fd);
 
         // this will emit repaintNeeded() as well when it is done
         d->updateAndSignalSizes();
@@ -130,7 +137,7 @@ void FrameSvg::setEnabledBorders(const EnabledBorders borders)
     fd->enabledBorders = oldBorders;
 
     //qDebug() << "looking for" << newKey;
-    FrameData *newFd = FrameSvgPrivate::s_sharedFrames.value(newKey);
+    FrameData *newFd = FrameSvgPrivate::s_sharedFrames[theme()->d].value(newKey);
     if (newFd) {
         //qDebug() << "FOUND IT!" << newFd->refcount;
         // we've found a math, so insert that new one and ref it ..
@@ -141,7 +148,7 @@ void FrameSvg::setEnabledBorders(const EnabledBorders borders)
         if (fd->deref(this)) {
             //const QString oldKey = d->cacheId(fd, d->prefix);
             //qDebug() << "1. Removing it" << oldKey << fd->refcount;
-            FrameSvgPrivate::s_sharedFrames.remove(oldKey);
+            FrameSvgPrivate::s_sharedFrames[theme()->d].remove(oldKey);
             delete fd;
         }
 
@@ -151,7 +158,7 @@ void FrameSvg::setEnabledBorders(const EnabledBorders borders)
     if (fd->refcount() == 1) {
         // we're the only user of it, let's remove it from the shared keys
         // we don't want to deref it, however, as we'll still be using it
-        FrameSvgPrivate::s_sharedFrames.remove(oldKey);
+        FrameSvgPrivate::s_sharedFrames[theme()->d].remove(oldKey);
     } else {
         // others are using it, but we wish to change its size. so deref it,
         // then create a copy of it (we're automatically ref'd via the ctor),
@@ -171,7 +178,7 @@ FrameSvg::EnabledBorders FrameSvg::enabledBorders() const
         return NoBorder;
     }
 
-    QHash<QString, FrameData*>::const_iterator it = d->frames.constFind(d->prefix);
+    QHash<QString, FrameData *>::const_iterator it = d->frames.constFind(d->prefix);
 
     if (it != d->frames.constEnd()) {
         return it.value()->enabledBorders;
@@ -183,21 +190,21 @@ FrameSvg::EnabledBorders FrameSvg::enabledBorders() const
 void FrameSvg::setElementPrefix(Plasma::Types::Location location)
 {
     switch (location) {
-        case Types::TopEdge:
-            setElementPrefix("north");
-            break;
-        case Types::BottomEdge:
-            setElementPrefix("south");
-            break;
-        case Types::LeftEdge:
-            setElementPrefix("west");
-            break;
-        case Types::RightEdge:
-            setElementPrefix("east");
-            break;
-        default:
-            setElementPrefix(QString());
-            break;
+    case Types::TopEdge:
+        setElementPrefix("north");
+        break;
+    case Types::BottomEdge:
+        setElementPrefix("south");
+        break;
+    case Types::LeftEdge:
+        setElementPrefix("west");
+        break;
+    case Types::RightEdge:
+        setElementPrefix("east");
+        break;
+    default:
+        setElementPrefix(QString());
+        break;
     }
 
     d->location = location;
@@ -226,7 +233,7 @@ void FrameSvg::setElementPrefix(const QString &prefix)
             FrameData *newFd = 0;
             if (!oldFrameData->frameSize.isEmpty()) {
                 const QString key = d->cacheId(oldFrameData, d->prefix);
-                newFd = FrameSvgPrivate::s_sharedFrames.value(key);
+                newFd = FrameSvgPrivate::s_sharedFrames[theme()->d].value(key);
             }
 
             // we need to put this in the cache if we didn't find it in the shared frames
@@ -246,12 +253,12 @@ void FrameSvg::setElementPrefix(const QString &prefix)
                 const QString key = d->cacheId(oldFrameData, d->prefix);
                 //qDebug() << this << "     1. inserting as" << key;
 
-                FrameSvgPrivate::s_sharedFrames.insert(key, newFd);
+                FrameSvgPrivate::s_sharedFrames[theme()->d].insert(key, newFd);
             }
         } else {
             // couldn't find anything useful, so we just create something here
             // we don't have a size for it yet, so don't bother trying to share it just yet
-            FrameData *newFd = new FrameData(this);
+            FrameData *newFd = new FrameData(this, d->prefix);
             d->frames.insert(d->prefix, newFd);
         }
 
@@ -263,7 +270,7 @@ void FrameSvg::setElementPrefix(const QString &prefix)
         if (oldFrameData) {
             if (oldFrameData->deref(this)) {
                 const QString oldKey = d->cacheId(oldFrameData, oldPrefix);
-                FrameSvgPrivate::s_sharedFrames.remove(oldKey);
+                FrameSvgPrivate::s_sharedFrames[theme()->d].remove(oldKey);
                 delete oldFrameData;
             }
         }
@@ -272,7 +279,7 @@ void FrameSvg::setElementPrefix(const QString &prefix)
     d->location = Types::Floating;
 }
 
-bool FrameSvg::hasElementPrefix(const QString & prefix) const
+bool FrameSvg::hasElementPrefix(const QString &prefix) const
 {
     //for now it simply checks if a center element exists,
     //because it could make sense for certain themes to not have all the elements
@@ -286,21 +293,21 @@ bool FrameSvg::hasElementPrefix(const QString & prefix) const
 bool FrameSvg::hasElementPrefix(Plasma::Types::Location location) const
 {
     switch (location) {
-        case Types::TopEdge:
-            return hasElementPrefix("north");
-            break;
-        case Types::BottomEdge:
-            return hasElementPrefix("south");
-            break;
-        case Types::LeftEdge:
-            return hasElementPrefix("west");
-            break;
-        case Types::RightEdge:
-            return hasElementPrefix("east");
-            break;
-        default:
-            return hasElementPrefix(QString());
-            break;
+    case Types::TopEdge:
+        return hasElementPrefix("north");
+        break;
+    case Types::BottomEdge:
+        return hasElementPrefix("south");
+        break;
+    case Types::LeftEdge:
+        return hasElementPrefix("west");
+        break;
+    case Types::RightEdge:
+        return hasElementPrefix("east");
+        break;
+    default:
+        return hasElementPrefix(QString());
+        break;
     }
 }
 
@@ -338,7 +345,7 @@ void FrameSvg::resizeFrame(const QSizeF &size)
     fd->frameSize = currentSize;
 
     //qDebug() << "looking for" << newKey;
-    FrameData *newFd = FrameSvgPrivate::s_sharedFrames.value(newKey);
+    FrameData *newFd = FrameSvgPrivate::s_sharedFrames[theme()->d].value(newKey);
     if (newFd) {
         //qDebug() << "FOUND IT!" << newFd->refcount;
         // we've found a math, so insert that new one and ref it ..
@@ -349,7 +356,7 @@ void FrameSvg::resizeFrame(const QSizeF &size)
         if (fd->deref(this)) {
             //const QString oldKey = d->cacheId(fd, d->prefix);
             //qDebug() << "1. Removing it" << oldKey << fd->refcount;
-            FrameSvgPrivate::s_sharedFrames.remove(oldKey);
+            FrameSvgPrivate::s_sharedFrames[theme()->d].remove(oldKey);
             delete fd;
         }
 
@@ -359,7 +366,7 @@ void FrameSvg::resizeFrame(const QSizeF &size)
     if (fd->refcount() == 1) {
         // we're the only user of it, let's remove it from the shared keys
         // we don't want to deref it, however, as we'll still be using it
-        FrameSvgPrivate::s_sharedFrames.remove(oldKey);
+        FrameSvgPrivate::s_sharedFrames[theme()->d].remove(oldKey);
     } else {
         // others are using it, but we wish to change its size. so deref it,
         // then create a copy of it (we're automatically ref'd via the ctor),
@@ -372,12 +379,12 @@ void FrameSvg::resizeFrame(const QSizeF &size)
     d->updateSizes();
     fd->frameSize = size.toSize();
     // we know it isn't in s_sharedFrames due to the check above, so insert it now
-    FrameSvgPrivate::s_sharedFrames.insert(newKey, fd);
+    FrameSvgPrivate::s_sharedFrames[theme()->d].insert(newKey, fd);
 }
 
 QSizeF FrameSvg::frameSize() const
 {
-    QHash<QString, FrameData*>::const_iterator it = d->frames.constFind(d->prefix);
+    QHash<QString, FrameData *>::const_iterator it = d->frames.constFind(d->prefix);
 
     if (it == d->frames.constEnd()) {
         return QSize(-1, -1);
@@ -395,20 +402,20 @@ qreal FrameSvg::marginSize(const Plasma::Types::MarginEdge edge) const
     switch (edge) {
     case Plasma::Types::TopMargin:
         return d->frames[d->prefix]->topMargin;
-    break;
+        break;
 
     case Plasma::Types::LeftMargin:
         return d->frames[d->prefix]->leftMargin;
-    break;
+        break;
 
     case Plasma::Types::RightMargin:
         return d->frames[d->prefix]->rightMargin;
-    break;
+        break;
 
     //Plasma::BottomMargin
     default:
         return d->frames[d->prefix]->bottomMargin;
-    break;
+        break;
     }
 }
 
@@ -421,20 +428,20 @@ qreal FrameSvg::fixedMarginSize(const Plasma::Types::MarginEdge edge) const
     switch (edge) {
     case Plasma::Types::TopMargin:
         return d->frames[d->prefix]->fixedTopMargin;
-    break;
+        break;
 
     case Plasma::Types::LeftMargin:
         return d->frames[d->prefix]->fixedLeftMargin;
-    break;
+        break;
 
     case Plasma::Types::RightMargin:
         return d->frames[d->prefix]->fixedRightMargin;
-    break;
+        break;
 
     //Plasma::BottomMargin
     default:
         return d->frames[d->prefix]->fixedBottomMargin;
-    break;
+        break;
     }
 }
 
@@ -485,7 +492,7 @@ QRectF FrameSvg::contentsRect() const
 
 QPixmap FrameSvg::alphaMask() const
 {
-    //FIXME: the distinction between overlay and 
+    //FIXME: the distinction between overlay and
     return d->alphaMask();
 }
 
@@ -522,14 +529,14 @@ void FrameSvg::clearCache()
     FrameData *frame = d->frames[d->prefix];
 
     // delete all the frames that aren't this one
-    QMutableHashIterator<QString, FrameData*> it(d->frames);
+    QMutableHashIterator<QString, FrameData *> it(d->frames);
     while (it.hasNext()) {
         FrameData *p = it.next().value();
         if (frame != p) {
             //TODO: should we clear from the Theme pixmap cache as well?
             if (p->deref(this)) {
                 const QString key = d->cacheId(p, it.key());
-                FrameSvgPrivate::s_sharedFrames.remove(key);
+                FrameSvgPrivate::s_sharedFrames[theme()->d].remove(key);
                 p->cachedBackground = QPixmap();
             }
 
@@ -596,10 +603,10 @@ FrameSvgPrivate::~FrameSvgPrivate()
                 const QString key = cacheId(it.value(), it.key());
 #ifdef DEBUG_FRAMESVG_CACHE
 #ifndef NDEBUG
-                // qDebug() << "2. Removing it" << key << it.value() << it.value()->refcount() << s_sharedFrames.contains(key);
+                // qDebug() << "2. Removing it" << key << it.value() << it.value()->refcount() << s_sharedFrames[theme()->d].contains(key);
 #endif
 #endif
-                s_sharedFrames.remove(key);
+                s_sharedFrames[q->theme()->d].remove(key);
                 delete it.value();
             }
 #ifdef DEBUG_FRAMESVG_CACHE
@@ -617,7 +624,7 @@ FrameSvgPrivate::~FrameSvgPrivate()
     }
 
 #ifdef DEBUG_FRAMESVG_CACHE
-    QHashIterator<QString, FrameData *> it2(s_sharedFrames);
+    QHashIterator<QString, FrameData *> it2(s_sharedFrames[theme()->d]);
     int shares = 0;
     while (it2.hasNext()) {
         it2.next();
@@ -632,14 +639,14 @@ FrameSvgPrivate::~FrameSvgPrivate()
 #endif
             foreach (FrameSvg *data, it2.value()->references.keys()) {
 #ifndef NDEBUG
-                qDebug()<< "            " << (void*)data << it2.value()->references[data];
+                qDebug() << "            " << (void *)data << it2.value()->references[data];
 #endif
             }
             shares += rc - 1;
         }
     }
 #ifndef NDEBUG
-    // qDebug() << "#####################################" << s_sharedFrames.count() << ", pixmaps saved:" << shares;
+    // qDebug() << "#####################################" << s_sharedFrames[theme()->d].count() << ", pixmaps saved:" << shares;
 #endif
 #endif
 
@@ -676,27 +683,29 @@ QPixmap FrameSvgPrivate::alphaMask()
             // see if we can find a suitable candidate in the shared frames
             // if successful, ref and insert, otherwise create a new one
             // and insert that into both the shared frames and our frames.
-            FrameData *maskFrame = s_sharedFrames.value(key);
+            FrameData *maskFrame = s_sharedFrames[q->theme()->d].value(key);
 
             if (maskFrame) {
                 maskFrame->ref(q);
             } else {
                 maskFrame = new FrameData(*frame, q);
-                s_sharedFrames.insert(key, maskFrame);
+                s_sharedFrames[q->theme()->d].insert(key, maskFrame);
             }
+            maskFrame->enabledBorders = frame->enabledBorders;
 
             frames.insert(prefix, maskFrame);
             updateSizes();
         }
 
         FrameData *maskFrame = frames[prefix];
+        maskFrame->enabledBorders = frame->enabledBorders;
         if (maskFrame->cachedBackground.isNull() || maskFrame->frameSize != frameSize(frame)) {
             const QString oldKey = cacheId(maskFrame, prefix);
             maskFrame->frameSize = frameSize(frame).toSize();
             const QString newKey = cacheId(maskFrame, prefix);
-            if (s_sharedFrames.contains(oldKey)) {
-                s_sharedFrames.remove(oldKey);
-                s_sharedFrames.insert(newKey, maskFrame);
+            if (s_sharedFrames[q->theme()->d].contains(oldKey)) {
+                s_sharedFrames[q->theme()->d].remove(oldKey);
+                s_sharedFrames[q->theme()->d].insert(newKey, maskFrame);
             }
 
             maskFrame->cachedBackground = QPixmap();
@@ -725,10 +734,10 @@ void FrameSvgPrivate::generateBackground(FrameData *frame)
     const bool overlayAvailable = !prefix.startsWith(QLatin1String("mask-")) && q->hasElement(prefix % "overlay");
     QPixmap overlay;
     if (q->isUsingRenderingCache()) {
-        frameCached = theme->findInCache(id, frame->cachedBackground) && !frame->cachedBackground.isNull();
+        frameCached = q->theme()->findInCache(id, frame->cachedBackground) && !frame->cachedBackground.isNull();
 
         if (overlayAvailable) {
-            overlayCached = theme->findInCache("overlay_" % id, overlay) && !overlay.isNull();
+            overlayCached = q->theme()->findInCache("overlay_" % id, overlay) && !overlay.isNull();
         }
     }
 
@@ -746,7 +755,7 @@ void FrameSvgPrivate::generateBackground(FrameData *frame)
             actualOverlayPos.setX(frame->frameSize.width() - overlaySize.width());
         } else if (q->hasElement(prefix % "hint-overlay-pos-bottom")) {
             actualOverlayPos.setY(frame->frameSize.height() - overlaySize.height());
-        //Stretched or Tiled?
+            //Stretched or Tiled?
         } else if (q->hasElement(prefix % "hint-overlay-stretch")) {
             overlaySize = frameSize(frame).toSize();
         } else {
@@ -763,12 +772,12 @@ void FrameSvgPrivate::generateBackground(FrameData *frame)
         overlayPainter.setCompositionMode(QPainter::CompositionMode_SourceIn);
         //Tiling?
         if (q->hasElement(prefix % "hint-overlay-tile-horizontal") ||
-            q->hasElement(prefix % "hint-overlay-tile-vertical")) {
+                q->hasElement(prefix % "hint-overlay-tile-vertical")) {
 
             QSize s = q->size();
             q->resize(q->elementSize(prefix % "overlay"));
 
-            overlayPainter.drawTiledPixmap(QRect(QPoint(0,0), overlaySize), q->pixmap(prefix % "overlay"));
+            overlayPainter.drawTiledPixmap(QRect(QPoint(0, 0), overlaySize), q->pixmap(prefix % "overlay"));
             q->resize(s);
         } else {
             q->paint(&overlayPainter, QRect(actualOverlayPos, overlaySize), prefix % "overlay");
@@ -797,7 +806,6 @@ void FrameSvgPrivate::generateFrameBackground(FrameData *frame)
     const int topOffset = 0;
     const int leftOffset = 0;
 
-
     if (!size.isValid()) {
 #ifndef NDEBUG
         // qDebug() << "Invalid frame size" << size;
@@ -848,11 +856,11 @@ void FrameSvgPrivate::generateFrameBackground(FrameData *frame)
         if (contentHeight > 0 && contentWidth > 0) {
             if (frame->composeOverBorder) {
                 q->paint(&p, QRect(QPoint(0, 0), size.toSize()),
-                               prefix % "center");
+                         prefix % "center");
             } else {
                 q->paint(&p, QRect(frame->leftWidth, frame->topHeight,
-                                contentWidth, contentHeight),
-                                prefix % "center");
+                                   contentWidth, contentHeight),
+                         prefix % "center");
             }
         }
     }
@@ -899,26 +907,26 @@ void FrameSvgPrivate::generateFrameBackground(FrameData *frame)
     if (frame->stretchBorders) {
         if (frame->enabledBorders & FrameSvg::LeftBorder || frame->enabledBorders & FrameSvg::RightBorder) {
             if (q->hasElement(prefix % "left") &&
-                frame->enabledBorders & FrameSvg::LeftBorder &&
-                contentHeight > 0) {
+                    frame->enabledBorders & FrameSvg::LeftBorder &&
+                    contentHeight > 0) {
                 q->paint(&p, QRect(leftOffset, contentTop, frame->leftWidth, contentHeight), prefix % "left");
             }
 
             if (q->hasElement(prefix % "right") &&
-                frame->enabledBorders & FrameSvg::RightBorder &&
-                contentHeight > 0) {
+                    frame->enabledBorders & FrameSvg::RightBorder &&
+                    contentHeight > 0) {
                 q->paint(&p, QRect(rightOffset, contentTop, frame->rightWidth, contentHeight), prefix % "right");
             }
         }
 
         if (frame->enabledBorders & FrameSvg::TopBorder || frame->enabledBorders & FrameSvg::BottomBorder) {
             if (frame->enabledBorders & FrameSvg::TopBorder && q->hasElement(prefix % "top") &&
-                contentWidth > 0) {
+                    contentWidth > 0) {
                 q->paint(&p, QRect(contentLeft, topOffset, contentWidth, frame->topHeight), prefix % "top");
             }
 
             if (frame->enabledBorders & FrameSvg::BottomBorder && q->hasElement(prefix % "bottom") &&
-                contentWidth > 0) {
+                    contentWidth > 0) {
                 q->paint(&p, QRect(contentLeft, bottomOffset, contentWidth, frame->bottomHeight), prefix % "bottom");
             }
         }
@@ -978,7 +986,7 @@ QString FrameSvgPrivate::cacheId(FrameData *frame, const QString &prefixToSave)
 {
     const QSize size = frameSize(frame).toSize();
     const QLatin1Char s('_');
-    return QString::number(frame->enabledBorders) % s % QString::number(size.width()) % s % QString::number(size.height()) % s % prefixToSave % s % q->imagePath(); 
+    return QString::number(frame->enabledBorders) % s % QString::number(size.width()) % s % QString::number(size.height()) % s % prefixToSave % s % q->imagePath();
 }
 
 void FrameSvgPrivate::cacheFrame(const QString &prefixToSave, const QPixmap &background, const QPixmap &overlay)
@@ -998,11 +1006,11 @@ void FrameSvgPrivate::cacheFrame(const QString &prefixToSave, const QPixmap &bac
 
     //qDebug()<<"Saving to cache frame"<<id;
 
-    theme->insertIntoCache(id, background, QString::number((qint64)q, 16) % prefixToSave);
+    q->theme()->insertIntoCache(id, background, QString::number((qint64)q, 16) % prefixToSave);
 
     if (!overlay.isNull()) {
         //insert overlay
-        theme->insertIntoCache("overlay_" % id, overlay, QString::number((qint64)q, 16) % prefixToSave % "overlay");
+        q->theme()->insertIntoCache("overlay_" % id, overlay, QString::number((qint64)q, 16) % prefixToSave % "overlay");
     }
 }
 
@@ -1025,7 +1033,6 @@ void FrameSvgPrivate::updateSizes() const
         frame->fixedTopMargin = frame->fixedTopHeight;
     }
 
-
     //The same, but its size depends from the margin being enabled
     if (frame->enabledBorders & FrameSvg::TopBorder) {
         frame->topHeight = q->elementSize(prefix % "top").height();
@@ -1167,5 +1174,4 @@ int FrameData::refcount() const
 
 } // Plasma namespace
 
-
 #include "moc_framesvg.cpp"
